# -*- coding: utf-8 -*-
"""
    emmett.helpers
    --------------

    Provides helping methods for applications.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import os
import re

from pydal.exceptions import NotAuthorizedException, NotFoundException

from .ctx import current, response, session
from .html import tag
from .http import HTTP, HTTPFile, HTTPIO

_re_dbstream = re.compile(r'(?P<table>.*?)\.(?P<field>.*?)\..*')


def abort(code, body=''):
    response.status = code
    raise HTTP(code, body)


def stream_file(path):
    full_path = os.path.join(current.app.root_path, path)
    raise HTTPFile(
        full_path, headers=response.headers, cookies=response.cookies
    )


def stream_dbfile(db, name):
    items = _re_dbstream.match(name)
    if not items:
        abort(404)
    table_name, field_name = items.group('table'), items.group('field')
    try:
        field = db[table_name][field_name]
    except AttributeError:
        abort(404)
    try:
        filename, path_or_stream = field.retrieve(name, nameonly=True)
    except NotAuthorizedException:
        abort(403)
    except NotFoundException:
        abort(404)
    except IOError:
        abort(404)
    if isinstance(path_or_stream, str):
        raise HTTPFile(
            path_or_stream, headers=response.headers, cookies=response.cookies
        )
    raise HTTPIO(
        path_or_stream, headers=response.headers, cookies=response.cookies
    )


def flash(message, category='message'):
    #: Flashes a message to the next request.
    if session._flashes is None:
        session._flashes = []
    session._flashes.append((category, message))


def get_flashed_messages(with_categories=False, category_filter=[]):
    #: Pulls flashed messages from the session and returns them.
    #  By default just the messages are returned, but when `with_categories`
    #  is set to `True`, the return value will be a list of tuples in the
    #  form `(category, message)` instead.
    if not isinstance(category_filter, list):
        category_filter = [category_filter]
    try:
        flashes = list(session._flashes or [])
        if category_filter:
            flashes = list(filter(lambda f: f[0] in category_filter, flashes))
        for el in flashes:
            session._flashes.remove(el)
        if not with_categories:
            return [x[1] for x in flashes]
    except Exception:
        flashes = []
    return flashes


def load_component(url, target=None, content='loading...'):
    attr = {}
    if target:
        attr['_id'] = target
    attr['_data-emt_remote'] = url
    return tag.div(content, **attr)



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.h11
    ------------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import ProtocolWrapper, protocols


@protocols.register('h11', packages=['h11'])
class H11Protocol(ProtocolWrapper):
    @classmethod
    def protocol_cls(cls):
        from uvicorn.protocols.http.h11_impl import H11Protocol
        return H11Protocol



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.auto
    -------------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import ProtocolWrapper, protocols


@protocols.register('auto')
class AutoProtocol(ProtocolWrapper):
    @classmethod
    def protocol_cls(cls):
        if 'httptools' in protocols.builders:
            return protocols.get_protocol('httptools').builder.protocol_cls()
        return protocols.get_protocol('h11').builder.protocol_cls()



# -*- coding: utf-8 -*-
"""
    emmett.asgi.loops
    -----------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..helpers import BuilderRegistry


class LoopRegistry(BuilderRegistry):
    def get_loop(self, name):
        try:
            builder, packages = self.builders[name]
        except KeyError:
            raise RuntimeError(
                'Event loop implementation "{}" not available'.format(name)
            )
        return builder(**packages)


loops = LoopRegistry()

from . import (
    asyncio,
    auto,
    uvloop
)



# -*- coding: utf-8 -*-
"""
    emmett.sessions
    ---------------

    Provides session managers for applications.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import os
import pickle
import tempfile
import time

from typing import Any, Optional, Type

from .ctx import current, request, response, websocket
from .datastructures import sdict, SessionData
from .pipeline import Pipe
from .security import secure_loads, secure_dumps, uuid
from .wrappers import ScopeWrapper


class SessionPipe(Pipe):
    def __init__(
        self, expire=3600, secure=False, domain=None, cookie_name=None
    ):
        self.expire = expire
        self.secure = secure
        self.domain = domain
        self.cookie_name = (
            cookie_name or f'emt_session_data_{current.app.name}'
        )

    def _load_session(self, wrapper: ScopeWrapper):
        raise NotImplementedError

    def _new_session(self) -> SessionData:
        raise NotImplementedError

    def _pack_session(self, expiration: int):
        response.cookies[self.cookie_name] = self._session_cookie_data()
        cookie_data = response.cookies[self.cookie_name]
        cookie_data['path'] = "/"
        cookie_data['expires'] = expiration
        if self.secure:
            cookie_data['secure'] = True
        if self.domain is not None:
            cookie_data['domain'] = self.domain

    def _session_cookie_data(self) -> str:
        raise NotImplementedError

    async def open_request(self):
        if self.cookie_name in request.cookies:
            current.session = self._load_session(request)
        if not current.session:
            current.session = self._new_session()

    async def open_ws(self):
        if self.cookie_name in websocket.cookies:
            current.session = self._load_session(websocket)
        if not current.session:
            current.session = self._new_session()

    async def close_request(self):
        expiration = current.session._expiration or self.expire
        self._pack_session(expiration)

    def clear(self):
        pass


class CookieSessionPipe(SessionPipe):
    def __init__(
        self, key, expire=3600, secure=False, domain=None, cookie_name=None
    ):
        super().__init__(expire, secure, domain, cookie_name)
        self.key = key

    def _load_session(self, wrapper: ScopeWrapper) -> SessionData:
        cookie_data = wrapper.cookies[self.cookie_name].value
        return SessionData(
            secure_loads(cookie_data, self.key),
            expires=self.expire
        )

    def _new_session(self) -> SessionData:
        return SessionData(expires=self.expire)

    def _session_cookie_data(self) -> str:
        return secure_dumps(sdict(current.session), self.key)

    def clear(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support sessions clearing. "
            f"Change the 'key' parameter to invalidate existing ones."
        )


class BackendStoredSessionPipe(SessionPipe):
    def _new_session(self):
        return SessionData(sid=uuid())

    def _session_cookie_data(self) -> str:
        return current.session._sid

    def _load_session(self, wrapper: ScopeWrapper) -> Optional[SessionData]:
        sid = wrapper.cookies[self.cookie_name].value
        data = self._load(sid)
        if data is not None:
            return SessionData(data, sid=sid)
        return None

    def _delete_session(self):
        pass

    def _save_session(self, expiration: int):
        pass

    def _load(self, sid: str):
        return None

    async def close_request(self):
        if not current.session:
            self._delete_session()
            if current.session._modified:
                #: if we got here means we want to destroy session definitely
                if self.cookie_name in response.cookies:
                    del response.cookies[self.cookie_name]
            return
        expiration = current.session._expiration or self.expire
        self._save_session(expiration)
        self._pack_session(expiration)


class FileSessionPipe(BackendStoredSessionPipe):
    _fs_transaction_suffix = '.__emt_sess'
    _fs_mode = 0o600

    def __init__(
        self, expire=3600, secure=False, domain=None, cookie_name=None,
        filename_template='emt_%s.sess'
    ):
        super().__init__(expire, secure, domain, cookie_name)
        assert not filename_template.endswith(self._fs_transaction_suffix), \
            'filename templates cannot end with %s' % \
            self._fs_transaction_suffix
        self._filename_template = filename_template
        self._path = os.path.join(current.app.root_path, 'sessions')
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)

    def _delete_session(self):
        fn = self._get_filename(current.session._sid)
        try:
            os.unlink(fn)
        except OSError:
            pass

    def _save_session(self, expiration):
        if current.session._modified:
            self._store(current.session, expiration)

    def _get_filename(self, sid):
        return os.path.join(self._path, self._filename_template % str(sid))

    def _load(self, sid):
        try:
            with open(self._get_filename(sid), 'rb') as f:
                exp = pickle.load(f)
                val = pickle.load(f)
        except IOError:
            return None
        if exp < time.time():
            return None
        return val

    def _store(self, session, expiration):
        fn = self._get_filename(session._sid)
        now = time.time()
        exp = now + expiration
        fd, tmp = tempfile.mkstemp(
            suffix=self._fs_transaction_suffix, dir=self._path)
        f = os.fdopen(fd, 'wb')
        try:
            pickle.dump(exp, f, 1)
            pickle.dump(sdict(session), f, pickle.HIGHEST_PROTOCOL)
        finally:
            f.close()
        try:
            os.rename(tmp, fn)
            os.chmod(fn, self._fs_mode)
        except Exception:
            pass

    def clear(self):
        for element in os.listdir(self._path):
            try:
                os.unlink(os.path.join(self._path, element))
            except Exception:
                pass


class RedisSessionPipe(BackendStoredSessionPipe):
    def __init__(
        self, redis, prefix="emtsess:", expire=3600, secure=False, domain=None,
        cookie_name=None
    ):
        super().__init__(expire, secure, domain, cookie_name)
        self.redis = redis
        self.prefix = prefix

    def _delete_session(self):
        self.redis.delete(self.prefix + current.session._sid)

    def _save_session(self, expiration):
        if current.session._modified:
            self.redis.setex(
                self.prefix + current.session._sid,
                expiration,
                current.session._dump
            )
        else:
            self.redis.expire(self.prefix + current.session._sid, expiration)

    def _load(self, sid):
        data = self.redis.get(self.prefix + sid)
        return pickle.loads(data) if data else data

    def clear(self):
        self.redis.delete(self.prefix + "*")


class SessionManager:
    _pipe: Optional[SessionPipe] = None

    @classmethod
    def _build_pipe(
        cls,
        handler_cls: Type[SessionPipe],
        *args,
        **kwargs
    ) -> SessionPipe:
        cls._pipe = handler_cls(*args, **kwargs)
        return cls._pipe

    @classmethod
    def cookies(
        cls,
        key: str,
        expire: int = 3600,
        secure: bool = False,
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None
    ) -> CookieSessionPipe:
        return cls._build_pipe(
            CookieSessionPipe,
            key,
            expire=expire,
            secure=secure,
            domain=domain,
            cookie_name=cookie_name
        )

    @classmethod
    def files(
        cls,
        expire: int = 3600,
        secure: bool = False,
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None,
        filename_template: str = 'emt_%s.sess'
    ) -> FileSessionPipe:
        return cls._build_pipe(
            FileSessionPipe,
            expire=expire,
            secure=secure,
            domain=domain,
            cookie_name=cookie_name,
            filename_template=filename_template
        )

    @classmethod
    def redis(
        cls,
        redis: Any,
        prefix: str = "emtsess:",
        expire: int = 3600,
        secure: bool = False,
        domain: Optional[str] = None,
        cookie_name: Optional[str] = None
    ) -> RedisSessionPipe:
        return cls._build_pipe(
            RedisSessionPipe,
            redis,
            prefix=prefix,
            expire=expire,
            secure=secure,
            domain=domain,
            cookie_name=cookie_name
        )

    @classmethod
    def clear(cls):
        cls._pipe.clear()



# -*- coding: utf-8 -*-
"""
    emmett.routing.router
    ---------------------

    Provides router implementations.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import re

from typing import Any, Callable, Dict, List, Type

from ..ctx import current
from ..http import HTTP
from .response import (
    ResponseBuilder, AutoResponseBuilder,
    BytesResponseBuilder, TemplateResponseBuilder
)
from .rules import RoutingRule, HTTPRoutingRule, WebsocketRoutingRule


class Router:
    __slots__ = [
        'app', 'routes_in', 'routes_out', '_routes_str', '_routes_nohost',
        '_get_routes_in_for_host', '_prefix_main', '_prefix_main_len',
        '_match_lang'
    ]

    _outputs: Dict[str, Type[ResponseBuilder]] = {}
    _routing_rule_cls: Type[RoutingRule] = RoutingRule
    _routing_signal = 'before_routes'
    _routing_started = False
    _routing_stack: List[RoutingRule] = []
    _re_components = re.compile(r'(\()?([^<\w]+)?<(\w+)\:(\w+)>(\)\?)?')

    def __init__(self, app, url_prefix=None):
        self.app = app
        self.routes_in = {'__any__': self._build_routing_dict()}
        self.routes_out = {}
        self._routes_str = {}
        self._routes_nohost = (self.routes_in['__any__'], )
        self._get_routes_in_for_host = self._get_routes_in_for_host_nomatch
        main_prefix = url_prefix or ''
        if main_prefix:
            main_prefix = main_prefix.rstrip('/')
            if not main_prefix.startswith('/'):
                main_prefix = '/' + main_prefix
            if main_prefix == '/':
                main_prefix = ''
        self._prefix_main = main_prefix
        self._prefix_main_len = len(self._prefix_main)
        self._set_language_handling()

    def _set_language_handling(self):
        self._match_lang = (
            self._match_with_lang if self.app.language_force_on_url
            else self._match_no_lang)

    @property
    def static_versioning(self):
        return (
            self.app.config.static_version_urls and
            self.app.config.static_version
        ) or ''

    @staticmethod
    def _build_routing_dict():
        return {'static': {}, 'match': {}}

    @classmethod
    def build_route_components(cls, path):
        components = []
        params = []
        for match in cls._re_components.findall(path):
            params.append(match[1] + "{}")
        statics = cls._re_components.sub("{}", path).split("{}")
        if not params:
            components = statics
        else:
            components.append(statics[0])
            for idx, el in enumerate(params):
                components.append(params[idx] + statics[idx + 1])
        return components

    def _get_routes_in_for_host_match(self, wrapper):
        return (
            self.routes_in.get(wrapper.host, self.routes_in['__any__']),
            self.routes_in['__any__'])

    def _get_routes_in_for_host_nomatch(self, wrapper):
        return self._routes_nohost

    def _match_with_lang(self, wrapper, path):
        path, lang = self._split_lang(path)
        current.language = wrapper.language = lang
        return path

    def _match_no_lang(self, wrapper, path):
        wrapper.language = None
        return path

    @staticmethod
    def remove_trailslash(path):
        return path.rstrip('/') or path

    def _split_lang(self, path):
        default = self.app.language_default
        if len(path) <= 1:
            return path, default
        clean_path = path.lstrip('/')
        if clean_path[2:3] == '/':
            lang, new_path = clean_path[:2], clean_path[2:]
            if lang != default and lang in self.app._languages_set:
                return new_path, lang
        return path, default

    def add_route(self, route):
        raise NotImplementedError

    def match(self, wrapper):
        raise NotImplementedError

    async def dispatch(self):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        if not self.__class__._routing_started:
            self.__class__._routing_started = True
            self.app.send_signal(self._routing_signal)
        return RoutingCtx(self, self._routing_rule_cls, *args, **kwargs)

    @classmethod
    def exposing(cls):
        return cls._routing_stack[-1]


class HTTPRouter(Router):
    __slots__ = ['pipeline', 'injectors']

    _routing_rule_cls = HTTPRoutingRule
    _outputs = {
        'auto': AutoResponseBuilder,
        'bytes': BytesResponseBuilder,
        'str': ResponseBuilder,
        'template': TemplateResponseBuilder
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = []
        self.injectors = []

    @staticmethod
    def _build_routing_dict():
        rv = {}
        for scheme in ['http', 'https']:
            rv[scheme] = {}
            for method in [
                'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'PATCH'
            ]:
                rv[scheme][method] = {'static': {}, 'match': {}}
        return rv

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s %s://%s%s%s -> %s" % (
            "|".join(route.methods).upper(),
            "|".join(route.schemes),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name
        )

    def add_route(self, route):
        host = route.hostname or '__any__'
        if host not in self.routes_in:
            self.routes_in[host] = self._build_routing_dict()
            self._get_routes_in_for_host = self._get_routes_in_for_host_match
        for scheme in route.schemes:
            for method in route.methods:
                routing_dict = self.routes_in[host][scheme][method.upper()]
                if route.is_static:
                    routing_dict['static'][route.path] = route
                else:
                    routing_dict['match'][route.name] = route
        self.routes_out[route.name] = {
            'host': route.hostname,
            'path': self.build_route_components(route.path)
        }
        self.add_route_str(route)

    def match(self, request):
        path = self.remove_trailslash(request.path)
        path = self._match_lang(request, path)
        for routing_dict in self._get_routes_in_for_host(request):
            sub_dict = routing_dict[request.scheme][request.method]
            route = sub_dict['static'].get(path)
            if route:
                return route, {}
            for route in sub_dict['match'].values():
                match, args = route.match(path)
                if match:
                    return route, args
        return None, {}

    async def dispatch(self):
        request, response = current.request, current.response
        route, reqargs = self.match(request)
        if not route:
            raise HTTP(404, body="Resource not found\n")
        request.name = route.name
        http_cls, output = await route.dispatcher.dispatch(request, reqargs)
        return http_cls(
            response.status, output, response.headers, response.cookies)


class WebsocketRouter(Router):
    __slots__ = ['pipeline']

    _routing_rule_cls = WebsocketRoutingRule

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = []

    @staticmethod
    def _build_routing_dict():
        rv = {}
        for scheme in ['ws', 'wss']:
            rv[scheme] = {'static': {}, 'match': {}}
        return rv

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s://%s%s%s -> %s" % (
            "|".join(route.schemes),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name
        )

    def add_route(self, route):
        host = route.hostname or '__any__'
        if host not in self.routes_in:
            self.routes_in[host] = self._build_routing_dict()
            self._get_routes_in_for_host = self._get_routes_in_for_host_match
        for scheme in route.schemes:
            routing_dict = self.routes_in[host][scheme]
            if route.is_static:
                routing_dict['static'][route.path] = route
            else:
                routing_dict['match'][route.name] = route
        self.routes_out[route.name] = {
            'host': route.hostname,
            'path': self.build_route_components(route.path)
        }
        self.add_route_str(route)

    def match(self, websocket):
        path = self.remove_trailslash(websocket.path)
        path = self._match_lang(websocket, path)
        for routing_dict in self._get_routes_in_for_host(websocket):
            sub_dict = routing_dict[websocket.scheme]
            route = sub_dict['static'].get(path)
            if route:
                return route, {}
            for route in sub_dict['match'].values():
                match, args = route.match(path)
                if match:
                    return route, args
        return None, {}

    async def dispatch(self):
        websocket = current.websocket
        route, reqargs = self.match(websocket)
        if not route:
            raise HTTP(404, body="Resource not found\n")
        websocket.name = route.name
        websocket._bind_flow(
            route.pipeline_flow_receive, route.pipeline_flow_send)
        await route.dispatcher.dispatch(websocket, reqargs)


class RoutingCtx:
    __slots__ = ['router', 'rule']

    def __init__(
        self,
        router: Router,
        rule_cls: Type[RoutingRule],
        *args,
        **kwargs
    ):
        self.router = router
        self.rule = rule_cls(self.router, *args, **kwargs)
        self.router._routing_stack.append(self.rule)

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        self.router.app.send_signal('before_route', route=self.rule, f=f)
        rv = self.rule(f)
        self.router.app.send_signal('after_route', route=self.rule)
        self.router._routing_stack.pop()
        return rv



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.request
    -----------------------

    Provides http request wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio
import pendulum
import re

from cgi import FieldStorage, parse_header
from datetime import datetime
from io import BytesIO
from urllib.parse import parse_qs
from typing import Optional

from ..asgi.typing import Scope, Receive, Send
from ..datastructures import sdict
from ..http import HTTP
from ..parsers import Parsers
from ..utils import cachedprop
from . import ScopeWrapper
from .helpers import FileStorage, RequestCancelled

_regex_client = re.compile(r'[\w\-:]+(\.[\w\-]+)*\.?')


class Body:
    __slots__ = ('_data', '_receive', '_max_content_length')

    def __init__(self, receive, max_content_length=None):
        self._data = bytearray()
        self._receive = receive
        self._max_content_length = max_content_length

    def append(self, data: bytes):
        if data == b'':
            return
        self._data.extend(data)
        if (
            self._max_content_length is not None and
            len(self._data) > self._max_content_length
        ):
            raise HTTP(413, 'Request entity too large')

    async def __load(self) -> bytes:
        while True:
            event = await self._receive()
            if event['type'] == 'http.request':
                self.append(event['body'])
                if not event.get('more_body', False):
                    break
            elif event['type'] == 'http.disconnect':
                raise RequestCancelled
        return bytes(self._data)

    def __await__(self):
        return self.__load().__await__()


class Request(ScopeWrapper):
    __slots__ = ['_now', 'method']

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        max_content_length: Optional[int] = None,
        body_timeout: Optional[int] = None
    ):
        super().__init__(scope, receive, send)
        self.max_content_length = max_content_length
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope['method']

    @cachedprop
    def _input(self):
        return Body(self._receive, self.max_content_length)

    @cachedprop
    async def body(self):
        if (
            self.max_content_length and
            self.content_length > self.max_content_length
        ):
            raise HTTP(413, 'Request entity too large')
        try:
            rv = await asyncio.wait_for(self._input, timeout=self.body_timeout)
        except asyncio.TimeoutError:
            raise HTTP(408, 'Request timeout')
        return rv

    @cachedprop
    def now(self):
        return pendulum.instance(self._now)

    @cachedprop
    def now_local(self):
        return self.now.in_timezone(pendulum.local_timezone())

    @cachedprop
    def content_type(self):
        return parse_header(self.headers.get('content-type', ''))[0]

    @cachedprop
    def content_length(self):
        return self.headers.get('content_length', 0, cast=int)

    _empty_body_methods = {v: v for v in ['GET', 'HEAD', 'OPTIONS']}

    @cachedprop
    async def _input_params(self):
        if self._empty_body_methods.get(self.method) or not self.content_type:
            return sdict(), sdict()
        return await self._load_params()

    @cachedprop
    async def body_params(self):
        rv, _ = await self._input_params
        return rv

    @cachedprop
    async def files(self):
        _, rv = await self._input_params
        return rv

    def _load_params_missing(self, data):
        return sdict(), sdict()

    def _load_params_json(self, data):
        try:
            params = Parsers.get_for('json')(data)
        except Exception:
            params = {}
        return sdict(params), sdict()

    def _load_params_form_urlencoded(self, data):
        rv = sdict()
        for key, values in parse_qs(
            data.decode(), keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv, sdict()

    def _load_params_form_multipart(self, data):
        params, files = sdict(), sdict()
        field_storage = FieldStorage(
            BytesIO(data),
            headers=self.headers,
            environ={'REQUEST_METHOD': self.method},
            keep_blank_values=True
        )
        for key in field_storage:
            field = field_storage[key]
            if isinstance(field, list):
                if len(field) > 1:
                    params[key] = []
                    for element in field:
                        params[key].append(element.value)
                else:
                    params[key] = field[0].value
            elif (
                isinstance(field, FieldStorage) and
                field.filename is not None
            ):
                files[key] = FileStorage(
                    BytesIO(field.file.read()),
                    field.filename,
                    field.name,
                    field.type,
                    field.headers
                )
                continue
            else:
                params[key] = field.value
        return params, files

    _params_loaders = {
        'application/json': _load_params_json,
        'application/x-www-form-urlencoded': _load_params_form_urlencoded,
        'multipart/form-data': _load_params_form_multipart
    }

    async def _load_params(self):
        loader = self._params_loaders.get(
            self.content_type, self._load_params_missing)
        return loader(self, await self.body)

    @cachedprop
    def client(self):
        g = _regex_client.search(self.headers.get('x-forwarded-for', ''))
        client = (g.group() or '').split(',')[0] if g else None
        if client in (None, '', 'unknown'):
            g = _regex_client.search(self.headers.get('remote-addr', ''))
            if g:
                client = g.group()
            elif self.host.startswith('['):
                # IPv6
                client = '::1'
            else:
                # IPv4
                client = '127.0.0.1'
        return client



# -*- coding: utf-8 -*-
"""
    emmett.extensions
    -----------------

    Provides base classes to create extensions.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from collections import OrderedDict
from functools import wraps


class listen_signal(object):
    _inst_count_ = 0

    def __init__(self, signal):
        if signal not in MetaExtension._signals_:
            raise SyntaxError('{} is not a valid signal'.format(signal))
        self.signal = signal
        self._inst_count_ = listen_signal._inst_count_
        listen_signal._inst_count_ += 1

    def __call__(self, f):
        self.f = f
        return self


class MetaExtension(type):
    _signals_ = [
        'before_routes',
        'before_route', 'after_route',
        'before_database', 'after_database',
        'after_loop'
    ]

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        declared_listeners = OrderedDict()
        all_listeners = OrderedDict()
        listeners = []
        for key, value in list(attrs.items()):
            if isinstance(value, listen_signal):
                listeners.append((key, value))
        listeners.sort(key=lambda x: x[1]._inst_count_)
        declared_listeners.update(listeners)
        new_class._declared_listeners_ = declared_listeners
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, '_declared_listeners_'):
                all_listeners.update(base._declared_listeners_)
        all_listeners.update(declared_listeners)
        new_class._all_listeners_ = all_listeners
        return new_class


class Extension(metaclass=MetaExtension):
    namespace = None
    default_config = {}

    def __init__(self, app, env, config):
        self.app = app
        self.env = env
        self.config = config
        self.__init_config()
        self.__init_listeners()

    def __init_config(self):
        for key, dval in self.default_config.items():
            self.config[key] = self.config.get(key, dval)

    def __init_listeners(self):
        self._listeners_ = []
        for name, obj in self._all_listeners_.items():
            self._listeners_.append((obj.signal, _wrap_listener(self, obj.f)))

    def on_load(self):
        pass


def _wrap_listener(ext, f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        return f(ext, *args, **kwargs)
    return wrapped



# -*- coding: utf-8 -*-
"""
    emmett.routing.dispatchers
    --------------------------

    Provides pipeline dispatchers for routes.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from ..ctx import current


class Dispatcher:
    __slots__ = ['f', 'flow_open', 'flow_close']

    def __init__(self, route):
        self.f = route.f
        self.flow_open = route.pipeline_flow_open
        self.flow_close = route.pipeline_flow_close

    async def _parallel_flow(self, flow):
        tasks = [asyncio.create_task(method()) for method in flow]
        await asyncio.gather(*tasks, return_exceptions=True)
        for task in tasks:
            if task.exception():
                raise task.exception()

    def dispatch(self, wrapper, reqargs):
        raise NotImplementedError


class RequestDispatcher(Dispatcher):
    __slots__ = ['response_builders']

    def __init__(self, route, rule):
        super().__init__(route)
        self.response_builders = rule.response_builders

    async def get_response(self, wrapper, reqargs):
        return self.response_builders[wrapper.method](await self.f(**reqargs))

    def dispatch(self, wrapper, reqargs):
        return self.get_response(wrapper, reqargs)


class RequestOpenDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        await self._parallel_flow(self.flow_open)
        return await self.get_response(wrapper, reqargs)


class RequestCloseDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        try:
            rv = await self.get_response(wrapper, reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return rv


class RequestFlowDispatcher(RequestDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        await self._parallel_flow(self.flow_open)
        try:
            rv = await self.get_response(wrapper, reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return rv


class WSDispatcher(Dispatcher):
    __slots__ = []

    def dispatch(self, wrapper, reqargs):
        return self.f(**reqargs)


class WSOpenDispatcher(WSDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        await self._parallel_flow(self.flow_open)
        await self.f(**reqargs)


class WSCloseDispatcher(WSDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        try:
            await self.f(**reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))


class WSFlowDispatcher(WSDispatcher):
    __slots__ = []

    async def dispatch(self, wrapper, reqargs):
        await self._parallel_flow(self.flow_open)
        try:
            await self.f(**reqargs)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await asyncio.shield(self._parallel_flow(self.flow_close))


class DispatcherCacheMixin:
    __slots__ = []
    _allowed_methods = {'GET', 'HEAD'}

    def __init__(self, route, rule):
        super().__init__(route, rule)
        self.route = route
        self.cache_rule = rule.cache_rule

    async def get_response(self, wrapper, reqargs):
        if wrapper.method not in self._allowed_methods:
            return await super().get_response(wrapper, reqargs)
        response = current.response
        key = self.cache_rule._build_ctx_key(
            self.route, **self.cache_rule._build_ctx(
                reqargs, self.route, current))
        data = self.cache_rule.cache.get(key)
        if data is not None:
            response.headers.update(data['headers'])
            return data['http_cls'], data['content']
        http_cls, output = await super().get_response(wrapper, reqargs)
        if response.status == 200:
            self.cache_rule.cache.set(
                key, {
                    'http_cls': http_cls,
                    'content': output,
                    'headers': response.headers},
                self.cache_rule.duration)
        return http_cls, output


class CacheDispatcher(DispatcherCacheMixin, RequestDispatcher):
    __slots__ = ('route', 'cache_rule')


class CacheOpenDispatcher(DispatcherCacheMixin, RequestOpenDispatcher):
    __slots__ = ('route', 'cache_rule')


class CacheCloseDispatcher(DispatcherCacheMixin, RequestCloseDispatcher):
    __slots__ = ('route', 'cache_rule')


class CacheFlowDispatcher(DispatcherCacheMixin, RequestFlowDispatcher):
    __slots__ = ('route', 'cache_rule')



# -*- coding: utf-8 -*-
"""
    emmett.serializers
    ------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from functools import partial
from rapidjson import (
    DM_ISO8601, DM_NAIVE_IS_UTC,
    NM_DECIMAL, NM_NATIVE,
    dumps as _json_dumps
)

from .html import tag, htmlescape

_json_safe_table = {
    'u2028': [r'\u2028', '\\u2028'],
    'u2029': [r'\u2029', '\\u2029']
}


class Serializers(object):
    _registry_ = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f
        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


def _json_default(obj):
    if hasattr(obj, '__json__'):
        return obj.__json__()
    raise ValueError('%r is not JSON serializable' % obj)


json = partial(
    _json_dumps,
    default=_json_default,
    datetime_mode=DM_ISO8601 | DM_NAIVE_IS_UTC,
    number_mode=NM_NATIVE | NM_DECIMAL
)


def json_safe(value):
    rv = json(value)
    for val, rep in _json_safe_table.values():
        rv.replace(val, rep)
    return rv


def xml_encode(value, key=None, quote=True):
    if hasattr(value, '__xml__'):
        return value.__xml__(key, quote)
    if isinstance(value, dict):
        return tag[key](
            *[
                tag[k](xml_encode(v, None, quote))
                for k, v in value.items()
            ])
    if isinstance(value, list):
        return tag[key](
            *[
                tag[item](xml_encode(item, None, quote))
                for item in value
            ])
    return htmlescape(value, quote)


Serializers.register_for('json')(json)


@Serializers.register_for('xml')
def xml(value, encoding='UTF-8', key='document', quote=True):
    rv = ('<?xml version="1.0" encoding="%s"?>' % encoding) + \
        str(xml_encode(value, key, quote))
    return rv



# -*- coding: utf-8 -*-
"""
    emmett.routing.rules
    --------------------

    Provides routing rules definition apis.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import os

from typing import Any, Callable

from ..cache import RouteCacheRule
from ..http import HTTPResponse
from ..pipeline import RequestPipeline, WebsocketPipeline, Pipe
from .routes import HTTPRoute, WebsocketRoute


class RoutingRule:
    __slots__ = ['router']

    def __init__(self, router, *args, **kwargs):
        self.router = router

    @property
    def app(self):
        return self.router.app

    def build_name(self, f):
        filename = os.path.realpath(f.__code__.co_filename)
        short = filename[1 + len(self.app.root_path):].rsplit('.', 1)[0]
        if not short:
            short = filename.rsplit('.', 1)[0]
        if short == "__init__":
            short = self.app.root_path.rsplit('/', 1)[-1]
        #: allow only one naming level if name is not provided
        if len(short.split(os.sep)) > 1:
            short = short.split(os.sep)[-1]
        return '.'.join(short.split(os.sep) + [f.__name__])

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        raise NotImplementedError


class HTTPRoutingRule(RoutingRule):
    __slots__ = [
        'name', 'f', 'output_type',
        'paths', 'schemes', 'methods', 'hostname', 'prefix',
        'template', 'template_folder', 'template_path',
        'pipeline', 'pipeline_flow_open', 'pipeline_flow_close',
        'response_builders', 'cache_rule'
    ]

    def __init__(
        self, router, paths=None, name=None, template=None, pipeline=None,
        injectors=None, schemes=None, hostname=None, methods=None, prefix=None,
        template_folder=None, template_path=None, cache=None, output='auto'
    ):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ('http', 'https')
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        self.methods = methods or ('get', 'post', 'head')
        if not isinstance(self.methods, (list, tuple)):
            self.methods = (self.methods,)
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith('/'):
                prefix = '/' + prefix
        self.prefix = prefix
        if output not in self.router._outputs:
            raise SyntaxError(
                'Invalid output specified. Allowed values are: {}'.format(
                    ', '.join(self.router._outputs.keys())))
        self.output_type = output
        self.template = template
        self.template_folder = template_folder
        self.template_path = template_path or self.app.template_path
        self.pipeline = (
            self.router.pipeline + (pipeline or []) +
            self.router.injectors + (injectors or []))
        self.cache_rule = None
        if cache:
            if not isinstance(cache, RouteCacheRule):
                raise RuntimeError(
                    'route cache argument should be a valid caching rule')
            if any(key in self.methods for key in ['get', 'head']):
                self.cache_rule = cache
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError('Invalid pipeline')

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        # is it good?
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        #
        if not self.template:
            self.template = f.__name__ + self.app.template_default_extension
        if self.template_folder:
            self.template = os.path.join(self.template_folder, self.template)
        pipeline_obj = RequestPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.f = wrapped_f
        output_type = pipeline_obj._output_type() or self.output_type
        self.response_builders = {
            method.upper(): self.router._outputs[output_type](self)
            for method in self.methods
        }
        if 'head' in self.response_builders:
            self.response_builders['head'].http_cls = HTTPResponse
        for idx, path in enumerate(self.paths):
            self.router.add_route(HTTPRoute(self, path, idx))
        return f


class WebsocketRoutingRule(RoutingRule):
    __slots__ = [
        'name', 'f',
        'paths', 'schemes', 'hostname', 'prefix',
        'pipeline', 'pipeline_flow_open', 'pipeline_flow_close',
        'pipeline_flow_receive', 'pipeline_flow_send'
    ]

    def __init__(
        self, router, paths=None, name=None, pipeline=None, schemes=None,
        hostname=None, prefix=None
    ):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ('ws', 'wss')
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith('/'):
                prefix = '/' + prefix
        self.prefix = prefix
        self.pipeline = self.router.pipeline + (pipeline or [])
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError('Invalid pipeline')

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        # is it good?
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        #
        pipeline_obj = WebsocketPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.pipeline_flow_receive = pipeline_obj._flow_receive()
        self.pipeline_flow_send = pipeline_obj._flow_send()
        self.f = wrapped_f
        for idx, path in enumerate(self.paths):
            self.router.add_route(WebsocketRoute(self, path, idx))
        return f



# -*- coding: utf-8 -*-
"""
    emmett.ctx
    ----------

    Provides the current object. Used by application to deal with
    request, response, session (if loaded with pipeline).

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import contextvars
import pendulum

from datetime import datetime

from ._internal import ContextVarProxy


class Context:
    def __init__(self):
        self.language = None

    @property
    def now(self):
        return pendulum.instance(datetime.utcnow())


class Current:
    __slots__ = ['_ctx']

    def __init__(self):
        object.__setattr__(self, '_ctx', contextvars.ContextVar('ctx'))
        self._ctx.set(Context())

    def _init_(self, ctx_cls, app, scope, receive, send, *args, **kwargs):
        return self._ctx.set(
            ctx_cls(app, scope, receive, send, *args, **kwargs)
        )

    def _close_(self, token):
        self._ctx.reset(token)

    @property
    def ctx(self):
        return self._ctx.get()

    def __getattr__(self, name):
        return getattr(self.ctx, name)

    def __setattr__(self, name, value):
        setattr(self.ctx, name, value)

    def __delattr__(self, name):
        delattr(self.ctx, name)

    @property
    def T(self):
        return self.ctx.app.translator


current = Current()
request = ContextVarProxy(current._ctx, 'request')
response = ContextVarProxy(current._ctx, 'response')
session = ContextVarProxy(current._ctx, 'session')
websocket = ContextVarProxy(current._ctx, 'websocket')


def now():
    return current.now



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols
    ---------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..helpers import BuilderRegistry


class ProtocolRegistry(BuilderRegistry):
    def get_protocol(self, name):
        try:
            builder, packages = self.builders[name]
        except KeyError:
            raise RuntimeError(
                'Porotocol "{}" not available'.format(name)
            )
        return _wrap_builder_with_packages(builder, packages)


class ProtocolWrapperRegistry(BuilderRegistry):
    def get_protocol(self, name):
        try:
            builder, packages = self.builders[name]
        except KeyError:
            raise RuntimeError(
                'Porotocol "{}" not available'.format(name)
            )
        return _wrap_builder_wrapper_with_packages(builder, packages)


# class Protocol(object):
#     def __init__(self, packages={}, *args, **kwargs):
#         for key, val in packages.items():
#             setattr(self, key, val)
#         self._init(*args, **kwargs)
#         self.init()

#     def _init(self, *args, **kwargs):
#         pass

#     def init(self):
#         pass


class ProtocolWrapper(object):
    def __init__(self, packages={}, *args, **kwargs):
        for key, val in packages.items():
            setattr(self, key, val)
        self._init(*args, **kwargs)

    def _init(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @classmethod
    def protocol_cls(cls):
        raise NotImplementedError

    def build_protocol(self):
        return self.protocol_cls()(*self._args, **self._kwargs)

    @classmethod
    def tick(cls):
        return cls.protocol_cls().tick()


def _wrap_builder_with_packages(builder, packages):
    def wrap(*args, **kwargs):
        return builder(packages=packages, *args, **kwargs)
    return wrap


class _wrap_builder_wrapper_with_packages(object):
    def __init__(self, builder, packages):
        self.builder = builder
        self.packages = packages

    def __call__(self, *args, **kwargs):
        return self.builder(
            packages=self.packages, *args, **kwargs).build_protocol()

    def tick(self):
        return self.builder.tick()


from .http import protocols as protocols_http
from .ws import protocols as protocols_ws



# -*- coding: utf-8 -*-
"""
    emmett._internal
    ----------------

    Provides internally used helpers and objects.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Flask and Werkzeug.
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import asyncio
import datetime
import os
import pendulum
import pkgutil
import sys
import warnings

from functools import partial
from shutil import copyfileobj


#: internal datastructures
class ObjectProxy(object):
    #: Proxy to another object.
    __slots__ = ('__obj', '__name__')

    def __init__(self, obj, name=None):
        object.__setattr__(self, '_ObjectProxy__obj', obj)
        object.__setattr__(self, '__name__', name)

    def _get_robj(self):
        try:
            return getattr(self.__obj, self.__name__)
        except AttributeError:
            raise RuntimeError('no object bound to %s' % self.__name__)

    @property
    def __dict__(self):
        try:
            return self._get_robj().__dict__
        except RuntimeError:
            raise AttributeError('__dict__')

    def __repr__(self):
        try:
            obj = self._get_robj()
        except RuntimeError:
            return '<%s unbound>' % self.__class__.__name__
        return repr(obj)

    def __bool__(self):
        try:
            return bool(self._get_robj())
        except RuntimeError:
            return False

    def __dir__(self):
        try:
            return dir(self._get_robj())
        except RuntimeError:
            return []

    def __getattr__(self, name):
        return getattr(self._get_robj(), name)

    def __setitem__(self, key, value):
        self._get_robj()[key] = value

    def __delitem__(self, key):
        del self._get_robj()[key]

    __setattr__ = lambda x, n, v: setattr(x._get_robj(), n, v)
    __delattr__ = lambda x, n: delattr(x._get_robj(), n)
    __str__ = lambda x: str(x._get_robj())
    __getitem__ = lambda x, i: x._get_robj()[i]
    __eq__ = lambda x, o: x._get_robj() == o
    __ne__ = lambda x, o: x._get_robj() != o
    __call__ = lambda x, *a, **kw: x._get_robj()(*a, **kw)
    __iter__ = lambda x: iter(x._get_robj())
    __contains__ = lambda x, i: i in x._get_robj()


class ContextVarProxy(ObjectProxy):
    __slots__ = ('__obj', '__name__')

    def __init__(self, obj, name=None):
        object.__setattr__(self, '_ContextVarProxy__obj', obj)
        object.__setattr__(self, '__name__', name)

    def _get_robj(self):
        return getattr(self.__obj.get(), self.__name__)


class ImmutableListMixin(object):
    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        _is_immutable(self)

    def __iadd__(self, other):
        _is_immutable(self)
    __imul__ = __iadd__

    def __setitem__(self, key, value):
        _is_immutable(self)

    def append(self, item):
        _is_immutable(self)
    remove = append

    def extend(self, iterable):
        _is_immutable(self)

    def insert(self, pos, value):
        _is_immutable(self)

    def pop(self, index=-1):
        _is_immutable(self)

    def reverse(self):
        _is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        _is_immutable(self)


class ImmutableList(ImmutableListMixin, list):
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__, list.__repr__(self)
        )


class LoopFileCtxWrapper:
    __slots__ = ('_coro', '_obj')

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class LoopFileWrapper:
    __slots__ = ('_file', '_loop')

    def __init__(self, f, loop=None):
        self._file = f
        self._loop = loop or asyncio.get_running_loop()

    async def read(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.read, *args, **kwargs))

    async def write(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.write, *args, **kwargs))

    async def close(self, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._file.close, *args, **kwargs))

    def __getattr__(self, name):
        return getattr(self._file, name)


def _is_immutable(self):
    raise TypeError('%r objects are immutable' % self.__class__.__name__)


async def _loop_open_file(loop, *args, **kwargs):
    f = await loop.run_in_executor(None, partial(open, *args, **kwargs))
    return LoopFileWrapper(f, loop)


def loop_open_file(*args, **kwargs):
    return LoopFileCtxWrapper(
        _loop_open_file(asyncio.get_running_loop(), *args, **kwargs)
    )


async def loop_copyfileobj(fsrc, fdst, length=None):
    return await asyncio.get_running_loop().run_in_executor(
        None, partial(copyfileobj, fsrc, fdst, length)
    )


#: deprecation helpers
class RemovedInNextVersionWarning(DeprecationWarning):
    pass


class deprecated(object):
    def __init__(self, old_method_name, new_method_name, class_name=None, s=0):
        self.class_name = class_name
        self.old_method_name = old_method_name
        self.new_method_name = new_method_name
        self.additional_stack = s

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            warn_of_deprecation(
                self.old_method_name, self.new_method_name, self.class_name,
                3 + self.additional_stack)
            return f(*args, **kwargs)
        return wrapped


warnings.simplefilter('always', RemovedInNextVersionWarning)


def warn_of_deprecation(old_name, new_name, prefix=None, stack=2):
    msg = "%(old)s is deprecated, use %(new)s instead."
    if prefix:
        msg = "%(prefix)s." + msg
    warnings.warn(
        msg % {'old': old_name, 'new': new_name, 'prefix': prefix},
        RemovedInNextVersionWarning, stack)


#: app init helpers
def get_root_path(import_name):
    """Returns the path of the package or cwd if that cannot be found."""
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, '__file__', None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError('No root path can be found for the provided '
                               'module "%s".  This can happen because the '
                               'module came from an import hook that does '
                               'not provide file name information or because '
                               'it\'s a namespace package.  In this case '
                               'the root path needs to be explicitly '
                               'provided.' % import_name)

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def create_missing_app_folders(app):
    try:
        for subfolder in ['languages', 'logs', 'static']:
            path = os.path.join(app.root_path, subfolder)
            if not os.path.exists(path):
                os.mkdir(path)
    except Exception:
        pass


#: monkey patches
def _pendulum_to_datetime(obj):
    return datetime.datetime(
        obj.year, obj.month, obj.day,
        obj.hour, obj.minute, obj.second, obj.microsecond,
        tzinfo=obj.tzinfo
    )


def _pendulum_to_naive_datetime(obj):
    obj = obj.in_timezone('UTC')
    return datetime.datetime(
        obj.year, obj.month, obj.day,
        obj.hour, obj.minute, obj.second, obj.microsecond
    )


pendulum.DateTime.as_datetime = _pendulum_to_datetime
pendulum.DateTime.as_naive_datetime = _pendulum_to_naive_datetime



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.ws.auto
    -----------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import ProtocolWrapper, protocols


@protocols.register('auto')
class AutoProtocol(ProtocolWrapper):
    @classmethod
    def protocol_cls(cls):
        if 'websockets' in protocols.builders:
            return protocols.get_protocol('websockets').builder.protocol_cls()
        return protocols.get_protocol('wsproto').builder.protocol_cls()



# -*- coding: utf-8 -*-
"""
    emmett.asgi.server
    ------------------

    Provides ASGI server wrapper over uvicorn.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import logging
import os

from uvicorn.config import Config as UvicornConfig, create_ssl_context
from uvicorn.lifespan.on import LifespanOn
from uvicorn.main import Server
from uvicorn.middleware.debug import DebugMiddleware
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware

from ..logger import LOG_LEVELS
from .loops import loops
from .protocols import protocols_http, protocols_ws


class Config(UvicornConfig):
    def setup_event_loop(self):
        pass

    def load(self):
        assert not self.loaded

        if self.is_ssl:
            self.ssl = create_ssl_context(
                keyfile=self.ssl_keyfile,
                certfile=self.ssl_certfile,
                ssl_version=self.ssl_version,
                cert_reqs=self.ssl_cert_reqs,
                ca_certs=self.ssl_ca_certs,
                ciphers=self.ssl_ciphers,
            )
        else:
            self.ssl = None

        encoded_headers = [
            (key.lower().encode("latin1"), value.encode("latin1"))
            for key, value in self.headers
        ]
        self.encoded_headers = (
            encoded_headers if b"server" in dict(encoded_headers) else
            [(b"server", b"Emmett")] + encoded_headers
        )

        self.http_protocol_class = self.http
        self.ws_protocol_class = self.ws
        self.lifespan_class = LifespanOn

        self.loaded_app = self.app
        self.interface = "asgi3"

        if self.debug:
            self.loaded_app = DebugMiddleware(self.loaded_app)
        if self.proxy_headers:
            self.loaded_app = ProxyHeadersMiddleware(
                self.loaded_app, trusted_hosts=self.forwarded_allow_ips
            )

        self.loaded = True


def run(
    app,
    host='127.0.0.1', port=8000, uds=None, fd=None,
    loop='auto', proto_http='auto', proto_ws='auto',
    log_level=None, access_log=None,
    proxy_headers=False, proxy_trust_ips=None,
    limit_concurrency=None,
    # limit_max_requests=None,
    backlog=2048,
    timeout_keep_alive=0
    # timeout_notify=30
):
    loop = loops.get_loop(loop)
    protocol_cls_http = protocols_http.get_protocol(proto_http)
    protocol_cls_ws = protocols_ws.get_protocol(proto_ws)

    app.send_signal('after_loop', loop=loop)

    if access_log is None:
        access_log = bool(app.debug)

    log_level = (
        LOG_LEVELS[log_level] if log_level else (
            logging.DEBUG if app.debug else logging.WARNING))

    if proxy_trust_ips is None:
        proxy_trust_ips = os.environ.get("PROXY_TRUST_IPS", "*")

    uvicorn_config = Config(
        app=app,
        host=host,
        port=port,
        uds=uds,
        fd=fd,
        loop=loop,
        http=protocol_cls_http,
        ws=protocol_cls_ws,
        log_level=log_level,
        access_log=access_log,
        debug=bool(app.debug),
        proxy_headers=proxy_headers,
        forwarded_allow_ips=proxy_trust_ips,
        limit_concurrency=limit_concurrency,
        # limit_max_requests=limit_max_requests,
        backlog=backlog,
        timeout_keep_alive=timeout_keep_alive,
        # timeout_notify=timeout_notify
    )
    server = Server(uvicorn_config)
    server.run()



# -*- coding: utf-8 -*-
"""
    emmett.parsers
    --------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from functools import partial
from rapidjson import (
    DM_ISO8601, DM_NAIVE_IS_UTC,
    NM_NATIVE,
    loads as _json_loads
)


class Parsers(object):
    _registry_ = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f
        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


json = partial(
    _json_loads,
    datetime_mode=DM_ISO8601 | DM_NAIVE_IS_UTC,
    number_mode=NM_NATIVE
)

Parsers.register_for('json')(json)



# -*- coding: utf-8 -*-
"""
    emmett.templating.templater
    ---------------------------

    Provides the Emmett implementation for Renoir engine.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from renoir import Renoir

from .lexers import lexers


class Templater(Renoir):
    def __init__(self, **kwargs):
        kwargs['lexers'] = lexers
        super().__init__(**kwargs)

    def _set_reload(self, value):
        self.cache.changes = value
        self.cache.load._configure()
        self.cache.prerender._configure()
        self.cache.parse._configure()

    def _set_encoding(self, value):
        self.encoding = value

    def _set_escape(self, value):
        self.escape = value
        self._configure()

    def _set_indent(self, value):
        self.indent = value
        self._configure()



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.ws.websockets
    -----------------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import ProtocolWrapper, protocols


@protocols.register('websockets', packages=['websockets'])
class WebsocketsProtocol(ProtocolWrapper):
    @classmethod
    def protocol_cls(cls):
        from uvicorn.protocols.websockets.websockets_impl import (
            WebSocketProtocol
        )
        return WebSocketProtocol



# -*- coding: utf-8 -*-
"""
    emmett.asgi.helpers
    -------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import sys


class BuilderRegistry(object):
    def __init__(self):
        self.registry = {}
        self.builders = {}

    def register(self, name, packages=[]):
        def wrap(builder):
            self.registry[name] = builder
            loaded_packages, implemented = {}, True
            try:
                for package in packages:
                    __import__(package)
                    loaded_packages[package] = sys.modules[package]
            except ImportError:
                implemented = False
            if implemented:
                self.builders[name] = (builder, loaded_packages)
            return builder
        return wrap



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.httptools
    ------------------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import ProtocolWrapper, protocols


@protocols.register('httptools', packages=['httptools'])
class HttpToolsProtocol(ProtocolWrapper):
    @classmethod
    def protocol_cls(cls):
        from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol
        return HttpToolsProtocol
